use super::{
    VerbContext,
    UserVerb,
    UserVerbRef,
    UResult,
    ItemSearchParams,
    user_error,
    get_player_item_or_fail,
    search_item_for_user,
};
use crate::{
    static_content::possession_type::possession_data,
    regular_tasks::queued_command::{
        QueueCommandHandler,
        QueueCommand,
        queue_command
    },
    models::item::{
        LocationActionType,
        SkillType,
    },
    services::{
        broadcast_to_room,
        skills::skill_check_and_grind,
    },
};
use async_trait::async_trait;
use std::time;

pub struct QueueHandler;
#[async_trait]
impl QueueCommandHandler for QueueHandler {
    async fn start_command(&self, ctx: &mut VerbContext<'_>, command: &QueueCommand)
                           -> UResult<time::Duration> {
        let player_item = get_player_item_or_fail(ctx).await?;
        if player_item.is_dead {
            user_error("You try to wield it, but your ghostly hands slip through it uselessly".to_owned())?;
        }
        let item_id = match command {
            QueueCommand::Wield { possession_id } => possession_id,
            _ => user_error("Unexpected command".to_owned())?
        };
        let item = match ctx.trans.find_item_by_type_code("possession", &item_id).await? {
            None => user_error("Item not found".to_owned())?,
            Some(it) => it
        };
        if item.location != format!("player/{}", player_item.item_code) {
            user_error("You try to wield it but realise you no longer have it".to_owned())?
        }
        let msg_exp = format!("{} fumbles around with {} {}\n",
                              &player_item.display_for_sentence(true, 1, true),
                              &player_item.pronouns.possessive,
                              &item.display_for_sentence(true, 1, false));
        let msg_nonexp = format!("{} fumbles around with {} {}\n",
                                 &player_item.display_for_sentence(false, 1, true),
                                 &player_item.pronouns.possessive,
                                 &item.display_for_sentence(false, 1, false));
        broadcast_to_room(ctx.trans, &player_item.location, None, &msg_exp, Some(&msg_nonexp)).await?;
        let mut draw_level: f64 = *player_item.total_skills.get(&SkillType::Quickdraw).to_owned().unwrap_or(&8.0);
        let mut player_item_mut = (*player_item).clone();
        
        let skill_result =
            skill_check_and_grind(ctx.trans, &mut player_item_mut, &SkillType::Quickdraw, draw_level).await?;
        if skill_result < -0.5 {
            draw_level -= 2.0;
        } else if skill_result < -0.25 {
            draw_level -= 1.0;
        } else if skill_result > 0.5 {
            draw_level += 2.0;
        } else if skill_result > 0.25 {
            draw_level += 1.0;
        }
        ctx.trans.save_item_model(&player_item_mut).await?;
        
        let wait_ticks = (12.0 - (draw_level / 2.0)).min(8.0).max(1.0);
        Ok(time::Duration::from_millis((wait_ticks * 500.0).round() as u64))
    }

    #[allow(unreachable_patterns)]
    async fn finish_command(&self, ctx: &mut VerbContext<'_>, command: &QueueCommand)
                            -> UResult<()> {
        let player_item = get_player_item_or_fail(ctx).await?;
        if player_item.is_dead {
            user_error("You try to wield it, but your ghostly hands slip through it uselessly".to_owned())?;
        }
        let item_id = match command {
            QueueCommand::Wield { possession_id } => possession_id,
            _ => user_error("Unexpected command".to_owned())?
        };
        let item = match ctx.trans.find_item_by_type_code("possession", &item_id).await? {
            None => user_error("Item not found".to_owned())?,
            Some(it) => it
        };
        if item.location != format!("player/{}", player_item.item_code) {
            user_error("You try to wield it but realise you no longer have it".to_owned())?
        }
        let msg_exp = format!("{} wields {} {}\n",
                              &player_item.display_for_sentence(true, 1, true),
                              &player_item.pronouns.possessive,
                              &item.display_for_sentence(true, 1, false));
        let msg_nonexp = format!("{} wields {} {}\n",
                                 &player_item.display_for_sentence(false, 1, true),
                                 &player_item.pronouns.possessive,
                                 &item.display_for_sentence(false, 1, false));
        broadcast_to_room(ctx.trans, &player_item.location, None, &msg_exp, Some(&msg_nonexp)).await?;
        ctx.trans.set_exclusive_action_type_to(&item,
                                               &LocationActionType::Wielded,
                                               &LocationActionType::Normal).await?;
        Ok(())
    }
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(self: &Self, ctx: &mut VerbContext, _verb: &str, remaining: &str) -> UResult<()> {
        let player_item = get_player_item_or_fail(ctx).await?;
        let weapon = search_item_for_user(ctx, &ItemSearchParams {
            include_contents: true,
            ..ItemSearchParams::base(&player_item, &remaining)
        }).await?;
        if player_item.is_dead {
            user_error("You try to wield it, but your ghostly hands slip through it uselessly".to_owned())?;
        }
        if weapon.action_type == LocationActionType::Wielded {
            user_error("You're actually already wielding it.".to_owned())?;
        }
        if weapon.item_type != "possession"  ||
            weapon.possession_type.as_ref()
            .and_then(|poss_type| possession_data().get(&poss_type))
            .and_then(|poss_data| poss_data.weapon_data.as_ref())
            .is_none() {
            user_error("You can't wield that!".to_owned())?;
        }
        queue_command(ctx, &QueueCommand::Wield { possession_id: weapon.item_code.clone() }).await?;
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
